#include <SoftwareSerial.h>
#include <Adafruit_NeoPixel.h>

extern "C" {
#include "user_interface.h"
}
// Version increment on change
#define SOFTWARE_NAME "CO²-Ampel"
#define SOFTWARE_VERSION "0.5"

/*****************************************************************
/* vars                                                          *
/*****************************************************************/
#define PIXEL_PIN   D8    // Digital IO pin connected to the NeoPixels.
#define PIXEL_COUNT 1
EspClass esp;
int CO2Wert;              // CO2 Messwert in ppm
uint32_t farbe = 0;
int Messinterval = 1;     // 1 Minute
char antwort[9];          // Buffer
//int TempOffset = 0;       // Sensor1 ZJAF10144
int TempOffset = 2;      // Sensor2 ZHC210025
int first = 0;
typedef struct {
  char flag[7];
  long count;
  int ppm;
} rtc;
rtc rtcMem;               // RTC memory

SoftwareSerial co2Serial(D6, D7); // RX, TX Pinns festlegen
Adafruit_NeoPixel strip = Adafruit_NeoPixel(PIXEL_COUNT, PIXEL_PIN, NEO_GRB + NEO_KHZ800);


/*****************************************************************
/* Commands MH-Z19B                                              *
/* muss noch optimaler programmiert werden                       *
/*****************************************************************/
// *  No Answer
// ** Answer 0x01, 0x00, ...
//byte cmd0x79[9] = {0xFF, 0x01, 0x79, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE6}; // Off Self-calibration for Zero Point **
byte cmd0x79[9] = {0xFF, 0x01, 0x79, 0xA0, 0x00, 0x00, 0x00, 0x00, 0x86}; // On Self-calibration for Zero Point **
byte cmd0x7d[9] = {0xFF, 0x01, 0x7D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x82}; // Self-calibration Status [r7] 
byte cmd0x86[9] = {0xFF, 0x01, 0x86, 0x00, 0x00, 0x00, 0x00, 0x00, 0x79}; // Read ppm [r2..3]. temp+40 [r4] 
byte cmd0x87[9] = {0xFF, 0x01, 0x87, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78}; // Zero point calibration *
byte cmd0x88[9] = {0xFF, 0x01, 0x88, 0x13, 0x88, 0x00, 0x00, 0x00, 0x64}; // Span point calibration *
byte cmd0x94[9] = {0xFF, 0x01, 0x94, 0x00, 0x00, 0x00, 0x13, 0x88, 0xCB}; // ID write
byte cmd0x95[9] = {0xFF, 0x01, 0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x6A}; // ID read
byte cmd0x99[9] = {0xFF, 0x01, 0x99, 0x00, 0x00, 0x00, 0x13, 0x88, 0xCB}; // Messbereich eintragen (5000) **
byte cmd0x9b[9] = {0xFF, 0x01, 0x9B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x64}; // Messbereich auslesen [r2..3] [r4..5]
byte cmd0xa0[9] = {0xFF, 0x01, 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5F}; // Firmware [r2..5]

/*****************************************************************
/* Display definitions                                           *
/*****************************************************************/
#define HAS_DISPLAY 1
#if (HAS_DISPLAY)
#include <SSD1306.h>
#include <SH1106.h>
#include "Font.h"                                 // Fonts for OLED

//SSD1306 display(0x3c, D2, D1);                    // (0x3x, SDA, SCL)
SSD1306 display(0x3d, D2, D1);                    // altenative Adresse
SH1106 display_sh1106(0x3c, D2, D1, GEOMETRY_128_32);
// OLED Display SSD1306 angeschlossen?
#define HAS_SSD1306 1
// OLED Display SH1106 angeschlossen?
#define HAS_SH1106  0
#define FLIP        1
#endif

void setup()
{
// Initialisierung des Displays              
#if (HAS_DISPLAY)
  if (HAS_SSD1306) {
    display.init();
    if (FLIP) display.flipScreenVertically();
  }
  if (HAS_SH1106) {
    display_sh1106.init();
    if (FLIP) display_sh1106.flipScreenVertically();
  }
#endif
  Serial.begin(115200);
//  Serial.begin(57600);
  Serial.println("");
  Serial.println("Start");
  first = LeseRTC();
  pinMode(D5, INPUT_PULLUP);
  strip.begin();                                  // WS2812B RGB initialisieren 

  co2Serial.begin(9600);
  if ( digitalRead(D5) == LOW ) Calibrate();      // Zero-Calbration
  leseBereich();                                  // MH-Z19 CO2 Sensor Arbeitsbereich
  if ( first ) {
    CO2Wert = leseCO2();                            // MH-Z19 CO2 Sensor lesen
    display_trend(CO2Wert - rtcMem.ppm);
  }
  leseID();                                       // Info über den Sensor
 
//  int temp = 150 - (150 * CO2Wert / 1800);        // CO2 in Farbe Umrechnen
//  if (temp < 0)temp += 256;
//  uint ADCValue = 0;
//  ADCValue = system_adc_read();                   // LDR Wert lesen
//  int x = ADCValue / 5;
//  if (x < 0)x = 0;
//  if (x > 100)x = 100;                      
//  farbe = Wheel(temp, x, x, x);                   // Farbe und Intensität einstellen

  int x = 20;                                     // in %
  TestLED();                                      // Debug. ggf. dealtivieren
  farbe = Wheel2(CO2Wert, x, x, x);               // Farbe nach Schwellwerten einstellen
  strip.setPixelColor(0, farbe);            
  strip.show();                                   // Daten zum WS2812B RGB LED senden

  SchreibeRTC();
  Serial.println("Ende");
  yield();
  esp.deepSleep(Messinterval * 60e6, WAKE_RF_DISABLED); // ESP8266 geht für x Minuten schlafen
  delay(200);                                     // geordnetes Deep Sleep
}
 
void loop()
{
  yield();
}

/*****************************************************************
/* RTC lesen                                                     *
/*****************************************************************/
int LeseRTC()
{
  int fehler = 0;
//  Serial.println();

  system_rtc_mem_read(64, &rtcMem, sizeof(rtcMem));
  if (rtcMem.flag[1] != 0) fehler = -1;
  if (rtcMem.flag[2] != 0) fehler = -1;
  if (rtcMem.flag[3] != 0) fehler = -1;
  if (rtcMem.flag[4] != 0xFF) fehler = -1;
  if (rtcMem.flag[5] != 0xFF) fehler = -1;
  if (rtcMem.flag[6] != 0xFF) fehler = -1;
  if (fehler) {
    rtcMem.flag[1] = 0;
    rtcMem.flag[2] = 0;
    rtcMem.flag[3] = 0;
    rtcMem.flag[4] = 0xFF;
    rtcMem.flag[5] = 0xFF;
    rtcMem.flag[6] = 0xFF;
    rtcMem.count = 0;
    display_init();
    delay(20000);
    Serial.println("Erster Start");
    return -1;
  }
  Serial.print("RTC Memory:  ");
  Serial.print(rtcMem.count, DEC);
  Serial.print("; vorher ");
  Serial.print(rtcMem.ppm, DEC);
  Serial.println("ppm");
}
      
/*****************************************************************
/* RTC lesen schreiben                                           *
/*****************************************************************/
void SchreibeRTC()
{
  rtcMem.count++;
  rtcMem.ppm = CO2Wert;
  system_rtc_mem_write(64,&rtcMem, sizeof(rtcMem));
}

/*****************************************************************
/* Ausgabe der ID                                                *
/*****************************************************************/
void leseID()
{
  Serial.print("Version : ");
  co2Serial.write(cmd0xa0, 9);
  co2Serial.readBytes(antwort, 9);
  int CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
  } else {
//  SerAnswer();
    Serial.print(char(antwort[2]));
    Serial.print(char(antwort[3]));
    Serial.print(char(antwort[4]));
    Serial.println(char(antwort[5]));
  }
  delay(20);

  Serial.print("ID      : ");
  cmd0x95[3] = 0x01;
  cmd0x95[8] = 0x69;
  co2Serial.write(cmd0x95, 9);
  co2Serial.readBytes(antwort, 9);
  CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
    return;
  } else {
//  SerAnswer();
    Serial.print(char(antwort[4]));
    Serial.print(char(antwort[5]));
    Serial.print(char(antwort[6]));
  }
  delay(20);

  cmd0x95[3] = 0x02;
  cmd0x95[8] = 0x68;
  co2Serial.write(cmd0x95, 9);
  co2Serial.readBytes(antwort, 9);
  CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
    return;
  } else {
//  SerAnswer();
    Serial.print(char(antwort[4]));
    Serial.print(char(antwort[5]));
    Serial.print(char(antwort[6]));
  }
  delay(20);

  cmd0x95[3] = 0x03;
  cmd0x95[8] = 0x67;
  co2Serial.write(cmd0x95, 9);
  co2Serial.readBytes(antwort, 9);
  CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
    return;
  } else {
//  SerAnswer();
    Serial.print(char(antwort[4]));
    Serial.print(char(antwort[5]));
    Serial.println(char(antwort[6]));
  }
}

/*****************************************************************
/* Ausgabe des calibrierten Messbereiches                        *
/*****************************************************************/
void leseBereich()
// Auslesen des Messbereiches des MH-Z19 CO2 Sensor
{
//  co2Serial.write(cmd0x79, 9);                    // Messraum eintragen
//  delay(200);
  co2Serial.write(cmd0x9b, 9);
  co2Serial.readBytes(antwort, 9);
  Serial.print("Messraum: ");
  int CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
    return;
  } else {
    int ppm = word(antwort[4], antwort[5]);
//  SerAnswer();
    Serial.print(ppm, DEC);
    Serial.println("ppm");
  }
}

/*****************************************************************
/* Zero-Calibrierung (auf 400ppm)                                *
/* Vor dieser muss der Sensor 20 Minuten in einer Umgebung       *
/* mit 400ppm laufen                                             *
/*****************************************************************/
// Anwort auswerten, mind. 20min überprüfen, Anzeige
void Calibrate()
{
  Serial.print  ("Wert    : ");
  Serial.println("Calbration: ");
  co2Serial.write(cmd0x87, 9);
}
  
/*****************************************************************
/* Auslesen des ppm-Wertes des MH-Z19 CO2 Sensors                *
/*****************************************************************/
int leseCO2()
{
  co2Serial.write(cmd0x86, 9);
  co2Serial.readBytes(antwort, 9);
  Serial.print("Wert    : ");
  int CS = Checksum();
  if (antwort[8] != CS) {
    Serial.println(" Fehler");
    return -1;
  } else {
//  SerAnswer();
  int ppm = word(antwort[2], antwort[3]);
  int temp1 = antwort[4] - 40 + TempOffset;       // Temeratur in °C ohne Nachkomma, [r4]-40 ungenau +/-2°?
    Serial.print(ppm, DEC);
    Serial.print("ppm; ");
    Serial.print(temp1, DEC);
    Serial.print("°C; ");
    Serial.print(antwort[6], DEC);
    Serial.print(", ");
    Serial.println(antwort[7], DEC);
  
#if defined(HAS_DISPLAY)
    display_oled(String(temp1, DEC), String(ppm, DEC));
#endif
    return ppm;                                     // Antwort des MH-Z19 CO2 Sensors in ppm
  }
}

/*****************************************************************
/* Checksum von anwort berechnen                                 *
/*****************************************************************/
int Checksum()
{
  int CS = 0xff -((antwort[0] + antwort[1] + antwort[2] + antwort[3] + antwort[4] + antwort[5] + antwort[6] + antwort[7]) % 0x100);
  return CS;
}

/*****************************************************************
/* Checksum von cmd berechnen                                 *
/*****************************************************************/
//int Checksum2()
//{
//  int CS = 0xff -((antwort[0] + antwort[1] + antwort[2] + antwort[3] + antwort[4] + antwort[5] + antwort[6] + antwort[7]) % 0x100);
//  return CS;
//}

/*****************************************************************
/* Durchlauf aller Farben (zum Test der Farben)                  *
/*****************************************************************/
void TestLED()
{
  int x = 20;
  farbe = Wheel2(200, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(500, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(700, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(1000, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(1300, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(1600, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);

  farbe = Wheel2(2000, x, x, x);
  strip.setPixelColor(0, farbe);
  strip.show();
  delay(250);
  
  strip.setPixelColor(0, 0);
  strip.show();
  delay(250);
}

/*****************************************************************
/* Input a value 0 to 255 to get a color value. alte Berechnung  *
/* The colours are a transition r - g - b - back to r.           *
/* alte Routine                                                  *
/*****************************************************************/
uint32_t Wheel(byte Wert, byte Rdimm, byte Gdimm, byte Bdimm)
{
  uint16_t Phase, Blue, Green, Red, Level;
  Phase = Wert / 42;
  Level = Wert % 42;
 
  switch ( Phase )
  {
    case 0:
      {
        Red   = 255;
        Green = Level * 6;
        Blue  = 0;
      }
      break;
 
    case 1:
      {
        Red   = 255 - Level * 6;
        Green = 255;
        Blue  = 0;
      }
      break;
 
    case 2:
      {
        Red   = 0;
        Green = 255;
        Blue  = Level * 6;
      }
      break;
 
    case 3:
      {
        Red   = 0;
        Green = 255 - Level * 6;
        Blue  = 255;
      }
      break;
 
    case 4:
      {
        Red   = Level * 6;
        Green = 0;
        Blue  = 255;
      }
      break;
    case 5:
      {
        Red   = 255;
        Green = 0;
        Blue  = 255 - Level * 6;
      }
      break;
 
    default:
      Red = Green = Blue = 255;
  }
  return strip.Color(Red * Rdimm / 100, Green * Gdimm / 100, Blue * Bdimm / 100);
  
}
 
/*****************************************************************
/* Input a ppm to get a color value                              *
/*****************************************************************/
uint32_t Wheel2(int Wert, byte Rdimm, byte Gdimm, byte Bdimm)
{
  uint16_t Phase, Blue, Green, Red, Level;
  if (Wert < 430)
  {
    Red   = 0;     // 0:    0 -  430
    Green = 0;
    Blue  = 10;
  } else {
    Red   = 0;     // 1:  430 -  650
    Green = 128;
    Blue  = 0;
  }
  if (Wert > 650)  // 2:  650 -  950
  {
    Red   = 0;
    Green = 255;
    Blue  = 0;
  }
  if (Wert > 950)  // 3:  950 - 1250
  {
    Red   = 128;
    Green = 255;
    Blue  = 0;
  }
  if (Wert > 1250) // 4: 1250 - 1500
  {
    Red   = 255;
    Green = 255;
    Blue  = 0;
  }
  if (Wert > 1500) // 5: 1500 - 1850
  {
    Red   = 255;
    Green = 128;
    Blue  = 0;
  }
  if (Wert > 1850) // 6: 1850 -
  {
    Red   = 255;
    Green = 0;
    Blue  = 0;
  }
  
  return strip.Color(Red * Rdimm / 100, Green * Gdimm / 100, Blue * Bdimm / 100);
}

/*****************************************************************
/* Serial Output (raw)                                           *
/*****************************************************************/
void SerAnswer()
{
  for ( int i = 0; i < 9; i++ ) {
    Serial.print(" ");
      if ( antwort[i] < 16 ) Serial.print("0");
    Serial.print(antwort[i], HEX);
  }
  Serial.print(": ");
}

/*****************************************************************
/* display values                                                *
/*****************************************************************/
void display_oled(const String& text1, const String& text2) {
#if (HAS_DISPLAY)
  int ofs = 0;
  if (FLIP) ofs = 16;
  if (HAS_SSD1306) {        // SSD1306 128x64 px für 0,66" 32-95 x 0-47
    display.clear();
    display.setFont(Roboto_Mono_9);
    display.displayOn();
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(32, ofs, "Temperatur:");
    display.drawString(32, 24 + ofs, "CO²:");
    display.drawString(78, 38 + ofs, "ppm");
    display.setFont(Roboto_Mono_14);
    display.drawString(82, 10 + ofs, "°C");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(74, 10 + ofs, text1);
    display.drawString(74, 34 + ofs, text2);
    display.display();
//    display.display();
  }
  if (HAS_SH1106) {         // SH1106 128x32 px
    display_sh1106.clear();
    display_sh1106.setFont(Roboto_Mono_14);
    display_sh1106.displayOn();
    display_sh1106.setTextAlignment(TEXT_ALIGN_LEFT);
    display_sh1106.drawString(0, 0, "Temp.:");
    display_sh1106.drawString(0, 14, "CO²:");
    display_sh1106.drawString(94, 0, "°C");
    display_sh1106.drawString(94, 15, "ppm");
    display_sh1106.setTextAlignment(TEXT_ALIGN_RIGHT);
    display_sh1106.drawString(90, 0, text1);
    display_sh1106.drawString(90, 15, text2);
    display_sh1106.display();
  }
#endif
}

/*****************************************************************
/* display init                                                *
/*****************************************************************/
void display_init() {
#if (HAS_DISPLAY)
  int ofs = 0;
  if (FLIP) ofs = 16;
  if (HAS_SSD1306) {        // SSD1306 128x64 px für 0,66" 32-95 x 0-47
    display.clear();
    display.displayOn();
    display.setFont(Roboto_Mono_9);
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(36, 8 + ofs, SOFTWARE_NAME);
    display.drawString(32, 30 + ofs, "Version:");
    display.drawString(80, 30 + ofs, SOFTWARE_VERSION);
    display.display();
  }
  if (HAS_SH1106) {         // SH1106 128x32 px
    display_sh1106.clear();
    display_sh1106.displayOn();
    display_sh1106.setFont(Roboto_Mono_14);
    display_sh1106.setTextAlignment(TEXT_ALIGN_LEFT);
    display_sh1106.drawString(28, 0, SOFTWARE_NAME);
    display.setFont(Roboto_Mono_9);
    display_sh1106.drawString(32, 15, "Version:");
    display_sh1106.drawString(80, 15, SOFTWARE_VERSION);
    display_sh1106.display();
  }
#endif
}

/*****************************************************************
/* display Trend                                                 *
/*****************************************************************/
void display_trend(int r)
{
#if (HAS_DISPLAY)
  int ofs = 0;
  if (FLIP) ofs = 16;
  if (HAS_SSD1306) {        // SSD1306 128x64; px für 0,66" 32-95 x 0-47
    if ( r > 16 ) {
      display.drawLine(82, 36 + ofs, 86, 28 + ofs);
      display.drawLine(90, 36 + ofs, 86, 28 + ofs);
      display.drawLine(82, 36 + ofs, 86, 30 + ofs);
      display.drawLine(90, 36 + ofs, 86, 30 + ofs);
    } else
    if ( r > 4 ) {
      display.drawLine(82, 36 + ofs, 86, 31 + ofs);
      display.drawLine(90, 36 + ofs, 86, 31 + ofs);
      display.drawLine(82, 36 + ofs, 86, 33 + ofs);
      display.drawLine(90, 36 + ofs, 86, 33 + ofs);
    }

    if ( r < -16 ) {
      display.drawLine(82, 28 + ofs, 86, 36 + ofs);
      display.drawLine(90, 28 + ofs, 86, 36 + ofs);
      display.drawLine(82, 28 + ofs, 86, 34 + ofs);
      display.drawLine(90, 28 + ofs, 86, 34 + ofs);
    } else
    if ( r < -4 ) {
      display.drawLine(82, 31 + ofs, 86, 36 + ofs);
      display.drawLine(90, 31 + ofs, 86, 36 + ofs);
      display.drawLine(82, 31 + ofs, 86, 34 + ofs);
      display.drawLine(90, 31 + ofs, 86, 34 + ofs);
    }
    display.display();
  }
#endif

}
